"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const winston = require("winston");
const aws_tags_1 = require("./aws-tags");
const aws_wrapper_1 = require("./aws-wrapper");
function getErrorsForFailedStack(stackId) {
    return __awaiter(this, void 0, void 0, function* () {
        const describeParams = {
            StackName: stackId
        };
        const describeResponse = yield aws_wrapper_1.default.cloudFormation.describeStackEvents(describeParams);
        const stackEvents = describeResponse.StackEvents;
        const failEvents = [];
        for (const stackEvent of stackEvents) {
            const resStatus = stackEvent.ResourceStatus;
            if (resStatus.includes('FAILED') && (stackEvent.ResourceStatusReason !== 'Resource creation cancelled' && stackEvent.ResourceStatusReason !== 'Resource update cancelled')) {
                failEvents.push(`${resStatus} : ${stackEvent.ResourceType} : ${stackEvent.ResourceStatusReason}`);
            }
            else if (stackEvent.ResourceType === 'AWS::CloudFormation::Stack' &&
                (resStatus === 'CREATE_COMPLETE' || resStatus === 'UPDATE_COMPLETE')) {
                break;
            }
        }
        return failEvents;
    });
}
function getCfErrorMessage(stackName, errors) {
    return `Errors while creating stack '${stackName}': \n${errors.join('\n')}`;
}
function delay(ms) {
    return __awaiter(this, void 0, void 0, function* () {
        yield new Promise(resolve => setTimeout(() => resolve(), ms)).then();
    });
}
/**
 * Given a stack name, returns the stack, or null if it doesnt exist
 */
function getStack(stackName) {
    return __awaiter(this, void 0, void 0, function* () {
        const params = {
            StackName: stackName
        };
        try {
            const describeResult = yield aws_wrapper_1.default.cloudFormation.describeStacks(params);
            return describeResult.Stacks[0];
        }
        catch (err) {
            if (err.code === 'ValidationError') { // Stack does not exist
                return null;
            }
            throw err;
        }
    });
}
exports.getStack = getStack;
/**
 * Waits for the given stack to be in the given state
 */
function waitForStack(stackName, stackState) {
    return __awaiter(this, void 0, void 0, function* () {
        const waitParams = {
            StackName: stackName
        };
        try {
            const waitResponse = yield aws_wrapper_1.default.cloudFormation.waitFor(stackState, waitParams);
            return waitResponse.Stacks[0];
        }
        catch (err) {
            if (err.originalError.code === 'Throttling') {
                winston.cli();
                winston.info(`Throttled. Sleeping ${err.retryDelay}ms`);
                yield delay(err.retryDelay);
                const waitResponse = yield aws_wrapper_1.default.cloudFormation.waitFor(stackState, waitParams);
                return waitResponse.Stacks[0];
            }
            throw (err);
        }
    });
}
exports.waitForStack = waitForStack;
/**
 * Given a stack name with template and parameters, creates the stack
 *
 * This method will wait for the stack to be in a create complete state
 * before resolving the promise
 */
function createStack(stackName, templateBodyOrUrl, parameters, timeoutInMinutes, tags) {
    return __awaiter(this, void 0, void 0, function* () {
        const params = {
            StackName: stackName,
            OnFailure: 'DELETE',
            Parameters: parameters,
            Capabilities: ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'],
            TimeoutInMinutes: timeoutInMinutes
        };
        if (templateBodyOrUrl.startsWith('s3://') || templateBodyOrUrl.startsWith('https://')) {
            params.TemplateURL = templateBodyOrUrl;
        }
        else {
            params.TemplateBody = templateBodyOrUrl;
        }
        // Add tags to the stack if provided
        if (tags) {
            params.Tags = aws_tags_1.toAWSTagStyle(tags);
        }
        const createResult = yield aws_wrapper_1.default.cloudFormation.createStack(params);
        const stackId = createResult.StackId;
        try {
            const createdStack = yield waitForStack(stackName, 'stackCreateComplete');
            return createdStack;
        }
        catch (err) {
            const errors = yield getErrorsForFailedStack(stackId);
            throw new Error(getCfErrorMessage(stackName, errors));
        }
    });
}
exports.createStack = createStack;
/**
 * Given a stack name with template and parameters, updates the existing stack
 *
 * This method will wait for the stack to be in an update complete state
 * before resolving the promise.
 */
function updateStack(stackName, templateBodyOrUrl, parameters, tags) {
    return __awaiter(this, void 0, void 0, function* () {
        const params = {
            StackName: stackName,
            Parameters: parameters,
            Capabilities: ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'],
        };
        if (templateBodyOrUrl.startsWith('s3://') || templateBodyOrUrl.startsWith('https://')) {
            params.TemplateURL = templateBodyOrUrl;
        }
        else {
            params.TemplateBody = templateBodyOrUrl;
        }
        // Add tags to the stack if provided
        if (tags) {
            params.Tags = aws_tags_1.toAWSTagStyle(tags);
        }
        try {
            const createResult = yield aws_wrapper_1.default.cloudFormation.updateStack(params);
            const stackId = createResult.StackId;
            try {
                const updatedStack = yield waitForStack(stackName, 'stackUpdateComplete');
                return updatedStack;
            }
            catch (err) {
                const errors = yield getErrorsForFailedStack(stackId);
                if (errors.length === 0) {
                    throw new Error('Error while waiting for stackUpdateComplete state on stack ' + stackName + ':\n ' + JSON.stringify(err));
                }
                throw new Error(getCfErrorMessage(stackName, errors));
            }
        }
        catch (err) {
            if (err.message.includes('No updates are to be performed')) {
                return exports.getStack(stackName);
            }
            throw err;
        }
    });
}
exports.updateStack = updateStack;
/**
 * Given a stack name, deletes the existing stack
 *
 * This method will wait for the stack to be in the delete complete state
 * before resolving the promise
 */
function deleteStack(stackName) {
    return __awaiter(this, void 0, void 0, function* () {
        const deleteParams = {
            StackName: stackName
        };
        const deleteResult = yield aws_wrapper_1.default.cloudFormation.deleteStack(deleteParams);
        const waitParams = {
            StackName: stackName
        };
        const waitResponse = yield aws_wrapper_1.default.cloudFormation.waitFor('stackDeleteComplete', waitParams);
        return true;
    });
}
exports.deleteStack = deleteStack;
/**
 * Given a regular JavaScript object of key/value pairs, returns the corresponding
 * CloudFormation parameters representation.
 *
 * Example:
 * The object {
 *   myparam: 'myvalue'
 * }
 *
 * ...becomes the object {
 *   ParameterKey: 'myparam',
 *   ParameterValue: 'myvalue',
 *   UsePreviousValue: false
 * }
 *
 * @param {Object} parametersObj - The object with plain key/value pairs to be changed into CF-style parameters
 * @returns {Object} - The object contianing CloudFormation-style parameters
 */
function getCfStyleStackParameters(parametersObj) {
    const stackParameters = [];
    for (const key in parametersObj) {
        if (parametersObj.hasOwnProperty(key)) {
            stackParameters.push({
                ParameterKey: key,
                ParameterValue: parametersObj[key],
                UsePreviousValue: false
            });
        }
    }
    return stackParameters;
}
exports.getCfStyleStackParameters = getCfStyleStackParameters;
/**
 * Given a CloudFormation stack, get the output for the given key
 *
 * @param {string} outputKey - The name of the output value to get
 * @param {Object} cfStack - The CF stack object returned from the CloudFormation API
 * @returns {string} - Returns either the output if it is present, or null if it isn't
 */
function getOutput(outputKey, cfStack) {
    for (const output of cfStack.Outputs) {
        if (output.OutputKey === outputKey) {
            return output.OutputValue;
        }
    }
    return null;
}
exports.getOutput = getOutput;
//# sourceMappingURL=cloudformation-calls.js.map