"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const aws_wrapper_1 = require("./aws-wrapper");
/**
 * Returns the information about the requested security group if it exists.
 */
function getSecurityGroup(groupName, vpcId) {
    return __awaiter(this, void 0, void 0, function* () {
        const describeSgParams = {
            Filters: [
                {
                    Name: 'vpc-id',
                    Values: [vpcId]
                },
                {
                    Name: 'group-name',
                    Values: [groupName]
                }
            ]
        };
        const describeResults = yield aws_wrapper_1.default.ec2.describeSecurityGroups(describeSgParams);
        if (describeResults.SecurityGroups && describeResults.SecurityGroups.length > 0) {
            return describeResults.SecurityGroups[0];
        }
        else {
            return null;
        }
    });
}
exports.getSecurityGroup = getSecurityGroup;
/**
 * Given the ID of a security group and VPC, returns the information about that
 * security group, or null if it doesn't exist.
 */
function getSecurityGroupById(groupId, vpcId) {
    return __awaiter(this, void 0, void 0, function* () {
        const describeSgParams = {
            Filters: [
                {
                    Name: 'vpc-id',
                    Values: [vpcId]
                },
                {
                    Name: 'group-id',
                    Values: [groupId]
                }
            ]
        };
        const describeResults = yield aws_wrapper_1.default.ec2.describeSecurityGroups(describeSgParams);
        if (describeResults.SecurityGroups && describeResults.SecurityGroups.length > 0) {
            return describeResults.SecurityGroups[0];
        }
        else {
            return null;
        }
    });
}
exports.getSecurityGroupById = getSecurityGroupById;
function removeIngressFromSg(sourceSg, ownSg, protocol, fromPort, toPort, vpcId) {
    return __awaiter(this, void 0, void 0, function* () {
        if (ingressRuleExists(ownSg, fromPort, toPort, protocol, sourceSg)) {
            const revokeParams = {
                GroupId: ownSg.GroupId,
                IpPermissions: [{
                        IpProtocol: protocol,
                        FromPort: fromPort,
                        ToPort: toPort,
                        UserIdGroupPairs: [
                            {
                                GroupId: sourceSg.GroupId,
                                VpcId: vpcId
                            }
                        ]
                    }]
            };
            yield aws_wrapper_1.default.ec2.revokeSecurityGroupIngress(revokeParams);
        }
        return true;
    });
}
exports.removeIngressFromSg = removeIngressFromSg;
/**
 * Given a security group, determines whether the given combination of protocol, start port,
 * end port, and source already exists as an ingress rule on the security group
 */
function ingressRuleExists(securityGroup, startPort, endPort, protocol, sourceSg) {
    let exists = false;
    for (const ingressRule of securityGroup.IpPermissions) {
        if (ingressRule.FromPort === startPort && ingressRule.ToPort === endPort && ingressRule.IpProtocol === protocol) {
            for (const ingressRuleSource of ingressRule.UserIdGroupPairs) {
                if (ingressRuleSource.GroupId === sourceSg.GroupId) {
                    exists = true;
                    break;
                }
            }
        }
    }
    return exists;
}
exports.ingressRuleExists = ingressRuleExists;
/**
 * Given a security group, adds an ingress rule from the given source security group
 * if that ingress rule doesn't already exist
 */
function addIngressRuleToSgIfNotExists(sourceSg, destSg, protocol, startPort, endPort, vpcId) {
    return __awaiter(this, void 0, void 0, function* () {
        const securityGroup = yield getSecurityGroup(destSg.GroupName, destSg.VpcId);
        if (securityGroup) {
            if (!ingressRuleExists(securityGroup, startPort, endPort, protocol, sourceSg)) {
                return addIngressRuleToSecurityGroup(sourceSg, destSg, protocol, startPort, endPort, vpcId);
            }
            else {
                return destSg;
            }
        }
        else {
            throw new Error('addIngressRuleToSgIfNotExists - missing security group');
        }
    });
}
exports.addIngressRuleToSgIfNotExists = addIngressRuleToSgIfNotExists;
/**
 * Adds an ingress rule from the given source security group to the given
 * destination security group
 */
function addIngressRuleToSecurityGroup(sourceSg, destSg, protocol, startPort, endPort, vpcId) {
    return __awaiter(this, void 0, void 0, function* () {
        const addIngressParams = {
            GroupId: destSg.GroupId,
            IpPermissions: [
                {
                    IpProtocol: protocol,
                    FromPort: startPort,
                    ToPort: endPort,
                    UserIdGroupPairs: [
                        {
                            GroupId: sourceSg.GroupId,
                            VpcId: vpcId
                        }
                    ]
                }
            ]
        };
        const authorizeResult = yield aws_wrapper_1.default.ec2.authorizeSecurityGroupIngress(addIngressParams);
        return getSecurityGroup(destSg.GroupName, vpcId);
    });
}
exports.addIngressRuleToSecurityGroup = addIngressRuleToSecurityGroup;
//# sourceMappingURL=ec2-calls.js.map