"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
/*
 * Copyright 2018 Brigham Young University
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
const fs = require("fs");
const awsTags = require("./aws-tags");
const aws_wrapper_1 = require("./aws-wrapper");
function getObjectsToDelete(objects) {
    const objectsToDelete = [];
    // Sort by date
    objects.sort((a, b) => {
        return b.LastModified.getTime() - a.LastModified.getTime();
    });
    // Always keep 5 most recent versions (no matter how old), then delete anything else older than 30 days
    if (objects.length > 5) {
        for (let i = 5; i < objects.length; i++) {
            const object = objects[i];
            const now = new Date();
            const diff = Math.abs(now.getTime() - object.LastModified.getTime());
            const daysOld = diff / (1000 * 60 * 60 * 24);
            if (daysOld > 30) {
                objectsToDelete.push(object);
            }
        }
    }
    return objectsToDelete;
}
/**
 * Delete all objects with a matching prefix from the given bucket
 */
function deleteMatchingPrefix(bucketName, prefix) {
    return __awaiter(this, void 0, void 0, function* () {
        const listObjectsParams = {
            Bucket: bucketName,
            Prefix: prefix
        };
        try {
            const listObjectsResponse = yield aws_wrapper_1.default.s3.listObjectsV2(listObjectsParams);
            const deleteObjectsParams = {
                Bucket: listObjectsResponse.Name,
                Delete: {
                    Objects: [],
                    Quiet: false
                }
            };
            for (const fl of listObjectsResponse.Contents) {
                deleteObjectsParams.Delete.Objects.push({
                    Key: fl.Key
                });
            }
            if (deleteObjectsParams.Delete.Objects.length === 0) {
                return null;
            }
            const deleteObjectsResponse = yield aws_wrapper_1.default.s3.deleteObjects(deleteObjectsParams);
            return deleteObjectsResponse;
        }
        catch (err) {
            // tslint:disable-next-line:no-console
            console.error(err);
            return null;
        }
    });
}
exports.deleteMatchingPrefix = deleteMatchingPrefix;
/**
 * Lists the files in a bucket for the given key prefix
 *
 * This method isn't sufficient currently for huge numbers of files under
 * a given prefix, since it only returns the first 1000 objects.
 * For Handel's purposes this is probably fine, but may need to be
 * extended in the future if we start getting more than 1000 objects
 * under a given key prefix
 */
function listFilesByPrefix(bucketName, keyPrefix) {
    return __awaiter(this, void 0, void 0, function* () {
        const listParams = {
            Bucket: bucketName,
            MaxKeys: 1000,
            Prefix: keyPrefix,
        };
        const listResponse = yield aws_wrapper_1.default.s3.listObjectsV2(listParams);
        let objects = [];
        if (listResponse.Contents) {
            objects = listResponse.Contents;
        }
        return objects;
    });
}
exports.listFilesByPrefix = listFilesByPrefix;
/**
 * Deletes the given objects from the S3 bucket
 */
function deleteFiles(bucketName, objects) {
    const objectsToDelete = [];
    for (const object of objects) {
        objectsToDelete.push({
            Key: object.Key
        });
    }
    const deleteParams = {
        Bucket: bucketName,
        Delete: {
            Objects: objectsToDelete,
            Quiet: true
        }
    };
    return aws_wrapper_1.default.s3.deleteObjects(deleteParams);
}
exports.deleteFiles = deleteFiles;
/**
 * Cleans up all objects in a bucket with a certain key prefix that are
 * older than 30 days.
 *
 * This method keeps at least the 5 most recent objects, no matter how
 * old they are.
 */
function cleanupOldVersionsOfFiles(bucketName, keyPrefix) {
    return __awaiter(this, void 0, void 0, function* () {
        const objects = yield listFilesByPrefix(bucketName, keyPrefix);
        const objectsToDelete = getObjectsToDelete(objects);
        if (objectsToDelete.length > 0) {
            return deleteFiles(bucketName, objectsToDelete);
        }
    });
}
exports.cleanupOldVersionsOfFiles = cleanupOldVersionsOfFiles;
/**
 * Uploads the given file to a bucket with the given key
 */
function uploadFile(bucketName, key, filePath) {
    return __awaiter(this, void 0, void 0, function* () {
        const fileStream = fs.createReadStream(filePath);
        const uploadParams = {
            Bucket: bucketName,
            Key: key,
            Body: fileStream
        };
        const uploadResponse = yield aws_wrapper_1.default.s3.upload(uploadParams);
        return uploadResponse;
    });
}
exports.uploadFile = uploadFile;
function uploadString(bucketName, key, content) {
    return __awaiter(this, void 0, void 0, function* () {
        const uploadParams = {
            Bucket: bucketName,
            Key: key,
            Body: Buffer.from(content, 'binary')
        };
        const uploadResponse = yield aws_wrapper_1.default.s3.upload(uploadParams);
        return uploadResponse;
    });
}
exports.uploadString = uploadString;
/**
 * Creates an S3 bucket in the given region
 */
function createBucket(bucketName, region, tags) {
    return __awaiter(this, void 0, void 0, function* () {
        const createParams = {
            Bucket: bucketName,
            ACL: 'private'
        };
        if (region !== 'us-east-1') { // If you specify us-east-1 it will fail (this is the default)
            createParams.CreateBucketConfiguration = {
                LocationConstraint: region
            };
        }
        const bucket = yield aws_wrapper_1.default.s3.createBucket(createParams);
        if (tags && Object.getOwnPropertyNames(tags).length > 0) {
            yield aws_wrapper_1.default.s3.putBucketTagging({
                Bucket: bucketName,
                Tagging: {
                    TagSet: awsTags.toAWSTagStyle(tags)
                }
            });
        }
        return bucket;
    });
}
exports.createBucket = createBucket;
/**
 * Gets information about the S3 bucket with the given name
 */
function getBucket(bucketName) {
    return __awaiter(this, void 0, void 0, function* () {
        const listResponse = yield aws_wrapper_1.default.s3.listBuckets();
        const buckets = listResponse.Buckets;
        for (const bucket of buckets) {
            if (bucket.Name === bucketName) {
                return bucket;
            }
        }
        return null;
    });
}
exports.getBucket = getBucket;
/**
 * Creates the S3 bucket with the given name and region, or just
 * returns the information about the bucket if it already exists
 */
function createBucketIfNotExists(bucketName, region, tags) {
    return __awaiter(this, void 0, void 0, function* () {
        const bucket = yield getBucket(bucketName);
        if (bucket) {
            return bucket;
        }
        else {
            const createResponse = yield createBucket(bucketName, region, tags || {});
            return getBucket(bucketName);
        }
    });
}
exports.createBucketIfNotExists = createBucketIfNotExists;
//# sourceMappingURL=s3-calls.js.map