"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
/*
 * Copyright 2018 Brigham Young University
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
const fs = require("fs");
const os = require("os");
const uuid = require("uuid");
const cloudFormationCalls = require("../aws/cloudformation-calls");
const s3Calls = require("../aws/s3-calls");
const ssmCalls = require("../aws/ssm-calls");
const util = require("../util/util");
// ------------------------------------------------------------------------------
// Public functions
// ------------------------------------------------------------------------------
function deployCloudFormationStack(serviceContext, stackName, cfTemplate, cfParameters, updatesSupported, timeoutInMinutes, stackTags) {
    return __awaiter(this, void 0, void 0, function* () {
        // Upload template
        const stack = yield cloudFormationCalls.getStack(stackName);
        if (!stack) {
            const s3ObjectData = yield uploadCFTemplateToHandelBucket(serviceContext, cfTemplate);
            return cloudFormationCalls.createStack(stackName, s3ObjectData.Location, cfParameters, timeoutInMinutes, stackTags);
        }
        else {
            if (updatesSupported) {
                const s3ObjectData = yield uploadCFTemplateToHandelBucket(serviceContext, cfTemplate);
                return cloudFormationCalls.updateStack(stackName, s3ObjectData.Location, cfParameters, stackTags);
            }
            else { // Updates not supported, so just return stack
                return stack;
            }
        }
    });
}
exports.deployCloudFormationStack = deployCloudFormationStack;
function getEnvVarsForDeployedService(ownServiceContext, dependenciesDeployContexts, userProvidedEnvVars) {
    let environmentVariables = {};
    // Inject env vars defined by service (if any)
    if (userProvidedEnvVars) {
        environmentVariables = Object.assign(environmentVariables, userProvidedEnvVars);
    }
    // Inject env vars defined by dependencies
    const dependenciesEnvVars = getEnvVarsFromDependencyDeployContexts(dependenciesDeployContexts);
    environmentVariables = Object.assign(environmentVariables, dependenciesEnvVars);
    // Inject env vars from Handel file
    const handelInjectedEnvVars = ownServiceContext.injectedEnvVars();
    environmentVariables = Object.assign(environmentVariables, handelInjectedEnvVars);
    return environmentVariables;
}
exports.getEnvVarsForDeployedService = getEnvVarsForDeployedService;
function getHandelUploadsBucketName(accountConfig) {
    return `handel-${accountConfig.region}-${accountConfig.account_id}`;
}
exports.getHandelUploadsBucketName = getHandelUploadsBucketName;
function uploadFileToHandelBucket(diskFilePath, artifactPrefix, s3FileName, accountConfig) {
    return __awaiter(this, void 0, void 0, function* () {
        const bucketName = yield ensureHandelBucketCreated(accountConfig);
        const artifactKey = `${artifactPrefix}/${s3FileName}`;
        const s3ObjectInfo = yield s3Calls.uploadFile(bucketName, artifactKey, diskFilePath);
        yield s3Calls.cleanupOldVersionsOfFiles(bucketName, artifactPrefix);
        return s3ObjectInfo;
    });
}
exports.uploadFileToHandelBucket = uploadFileToHandelBucket;
function uploadStringToHandelBucket(content, artifactPrefix, s3FileName, accountConfig) {
    return __awaiter(this, void 0, void 0, function* () {
        const bucketName = yield ensureHandelBucketCreated(accountConfig);
        const artifactKey = `${artifactPrefix}/${s3FileName}`;
        const s3ObjectInfo = yield s3Calls.uploadString(bucketName, artifactKey, content);
        yield s3Calls.cleanupOldVersionsOfFiles(bucketName, artifactPrefix);
        return s3ObjectInfo;
    });
}
exports.uploadStringToHandelBucket = uploadStringToHandelBucket;
function uploadDirectoryToHandelBucket(directoryPath, artifactPrefix, s3FileName, accountConfig) {
    return __awaiter(this, void 0, void 0, function* () {
        const zippedPath = `${os.tmpdir()}/${uuid()}`;
        yield util.zipDirectoryToFile(directoryPath, zippedPath);
        const s3ObjectInfo = yield uploadFileToHandelBucket(zippedPath, artifactPrefix, s3FileName, accountConfig);
        fs.unlinkSync(zippedPath); // Delete temporary file
        return s3ObjectInfo;
    });
}
exports.uploadDirectoryToHandelBucket = uploadDirectoryToHandelBucket;
function uploadCFTemplateToHandelBucket(serviceContext, templateBody) {
    return __awaiter(this, void 0, void 0, function* () {
        const accountConfig = serviceContext.accountConfig;
        const prefix = getServiceUploadLocation(serviceContext);
        const filename = `cfTemplates/template-${uuid()}`;
        return uploadStringToHandelBucket(templateBody, prefix, filename, accountConfig);
    });
}
exports.uploadCFTemplateToHandelBucket = uploadCFTemplateToHandelBucket;
function uploadDeployableArtifactToHandelBucket(serviceContext, pathToArtifact, s3FileName) {
    return __awaiter(this, void 0, void 0, function* () {
        const accountConfig = serviceContext.accountConfig;
        const fileStats = fs.lstatSync(pathToArtifact);
        const artifactPrefix = getServiceUploadLocation(serviceContext);
        if (fileStats.isDirectory()) { // Zip up artifact and upload it
            return uploadDirectoryToHandelBucket(pathToArtifact, artifactPrefix, s3FileName, accountConfig);
        }
        else { // Is file (i.e. WAR file or some other already-compiled archive), just upload directly
            return uploadFileToHandelBucket(pathToArtifact, artifactPrefix, s3FileName, accountConfig);
        }
    });
}
exports.uploadDeployableArtifactToHandelBucket = uploadDeployableArtifactToHandelBucket;
function getAllPolicyStatementsForServiceRole(serviceContext, ownServicePolicyStatements, dependenciesDeployContexts, includeAppSecretsStatements, includePutMetricStatements = false) {
    const policyStatementsToConsume = [];
    // Add policies from dependencies that have them
    for (const deployContext of dependenciesDeployContexts) {
        for (const policyDoc of deployContext.policies) {
            policyStatementsToConsume.push(policyDoc);
        }
    }
    // Let consuming service add its own policy if needed
    for (const ownServicePolicyStatement of ownServicePolicyStatements) {
        policyStatementsToConsume.push(ownServicePolicyStatement);
    }
    if (includeAppSecretsStatements) {
        const applicationParameters = `arn:aws:ssm:${serviceContext.accountConfig.region}:${serviceContext.accountConfig.account_id}:parameter/${serviceContext.appName}.${serviceContext.environmentName}*`;
        const applicationParametersPath = `arn:aws:ssm:${serviceContext.accountConfig.region}:${serviceContext.accountConfig.account_id}:parameter/${serviceContext.appName}/${serviceContext.environmentName}/*`;
        const appSecretsAcessStatements = [
            {
                Effect: 'Allow',
                Action: [
                    'ssm:DescribeParameters'
                ],
                Resource: [
                    '*'
                ]
            },
            {
                Effect: 'Allow',
                Action: [
                    'ssm:GetParameters',
                    'ssm:GetParameter',
                    'ssm:GetParametersByPath'
                ],
                Resource: [
                    applicationParameters,
                    applicationParametersPath,
                    `arn:aws:ssm:${serviceContext.accountConfig.region}:${serviceContext.accountConfig.account_id}:parameter/handel.global*`
                ]
            },
            {
                Effect: 'Allow',
                Action: [
                    'ssm:PutParameter',
                    'ssm:DeleteParameter',
                    'ssm:DeleteParameters'
                ],
                Resource: [
                    applicationParameters,
                    applicationParametersPath
                ]
            }
        ];
        policyStatementsToConsume.push(...appSecretsAcessStatements);
    }
    if (includePutMetricStatements) {
        const putMetricStatements = [
            {
                Effect: 'Allow',
                Action: [
                    'cloudwatch:PutMetricData'
                ],
                Resource: [
                    '*' // CloudWatch only allows '*' for metric permissions 🤷
                ]
            }
        ];
        policyStatementsToConsume.push(...putMetricStatements);
    }
    return policyStatementsToConsume;
}
exports.getAllPolicyStatementsForServiceRole = getAllPolicyStatementsForServiceRole;
function addItemToSSMParameterStore(ownServiceContext, paramName, paramValue) {
    return __awaiter(this, void 0, void 0, function* () {
        const fullParamName = ownServiceContext.ssmParamName(paramName);
        yield ssmCalls.storeParameter(fullParamName, 'SecureString', paramValue);
        return true;
    });
}
exports.addItemToSSMParameterStore = addItemToSSMParameterStore;
// ------------------------------------------------------------------------------
// Private functions
// ------------------------------------------------------------------------------
function getEnvVarsFromDependencyDeployContexts(deployContexts) {
    const envVars = {};
    for (const deployContext of deployContexts) {
        for (const envVarKey in deployContext.environmentVariables) {
            if (deployContext.environmentVariables.hasOwnProperty(envVarKey)) {
                envVars[envVarKey] = deployContext.environmentVariables[envVarKey];
            }
        }
    }
    return envVars;
}
function ensureHandelBucketCreated(accountConfig) {
    return __awaiter(this, void 0, void 0, function* () {
        const bucketName = getHandelUploadsBucketName(accountConfig);
        yield s3Calls.createBucketIfNotExists(bucketName, accountConfig.region, accountConfig.handel_resource_tags); // Ensure Handel bucket exists in this region
        return bucketName;
    });
}
function getServiceUploadLocation(serviceContext) {
    return `${serviceContext.appName}/${serviceContext.environmentName}/${serviceContext.serviceName}`;
}
//# sourceMappingURL=deploy-phase.js.map